"""
    Selective.PropOptimizerEGH.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize, basinhopping
from time import sleep
from molass_legacy.Models.ElutionCurveModels import egh, egha
from molass_legacy.QuickAnalysis.ModeledPeaks import recognize_peaks
from .PropOptimizerImpl import PROP_MIN_VALUE, AFFINE

egh_impl = egha if AFFINE else egh

def guess_init_params(x, y):
    peaks = recognize_peaks(x, y, num_peaks=2, affine=AFFINE)
    return np.asarray(peaks)

def compute_props(x, peaks):
    areas = []
    for params in peaks:
        cy = egh_impl(x, *params)
        areas.append(np.sum(cy))
    props = np.array(areas)/np.sum(areas)
    return props

def compute_cy_list(x, peaks):
    cy_list = []
    for params in peaks:
        cy = egh_impl(x, *params)
        cy_list.append(cy)
    return cy_list

def optimize_to_props(x, y, peaks, props, use_basinhopping=False):
    assert len(peaks) == len(props)
    peaks = np.asarray(peaks)
    props = np.asarray(props)

    def objective(p):
        ty = np.zeros(len(y))
        p_ = p.reshape(peaks.shape)
        areas = []
        try:
            cy_list = []
            for params in p_:
                cy = egh_impl(x, *params)
                cy_list.append(cy)
                areas.append(np.sum(cy))
                ty += cy
            props_ = np.array(areas)/np.sum(areas)
            area_dev = np.sum((ty - y)**2)
            prop_dev = max(PROP_MIN_VALUE, np.sum((props_ - props)**2))
            fv = np.log(area_dev) + np.log(prop_dev) - np.log(np.std(p_[:,1]))
        except:
            fv = np.inf
        return fv

    init_params = peaks.flatten()
    if use_basinhopping:
        ret = basinhopping(objective, init_params, minimizer_kwargs=dict(method="Nelder-Mead"))
    else:
        ret = minimize(objective, init_params, method="Nelder-Mead")
    return ret

def compute_optimal_proportion(progress_queue, job_args):
    for i in range(10):
        print([i])
        sleep(1)
        progress_queue.put([i])
    progress_queue.put([-1])



# coding: utf-8
"""
    DevelEnv.py

    Copyright (c) 2020-2022, Masatsuyo Takahashi, KEK-PF
"""
import ctypes
assert ctypes.windll.shell32.IsUserAnAdmin()
import sys
import os
import platform
from StdoutLogger import StdoutLogger
from .Embeddables import Embeddables

class DevelEnv(Embeddables):
    def __init__(self):
        self.logger = StdoutLogger("build.log")
        self.python_exe = sys.executable
        self.python00 = 'python' + ''.join(platform.python_version().split('.')[0:2])
        self.emb_folder, _ = os.path.split(self.python_exe)
        self.site_folder = os.path.join(self.emb_folder, r'Lib\site-packages')
        self.install_required_minimum()

    def build(self, stable=True):
        self.prepare_unofficial_site()
        self.install_requirements("requirements.txt")



# -*- coding: utf-8 -*-
"""

    ファイル名：   Build.py

    処理内容：

        配布形式のビルド

    Copyright (c) 2016-2018, Masatsuyo Takahashi, KEK-PF

"""

from __future__ import division, print_function, unicode_literals
import sys
import os
import re
import glob
import platform
import shutil
from subprocess         import call
from RequiredPackages   import check_tested
from BasicUtils         import clear_dirs_with_retry, rename_with_retry, mkdirs_with_retry

class_name_dict = {
    'synthesizer'       :   'Synthesizer',
    'autorg_kek'        :   'AutorgKek',
    'serial_analyzer'   :   'SerialAnalyzer',
    'pyinstaller-test'  :   'pyinstaller-test',
    'window_size_tester'    :   'SerialAnalyzer',
    'our_python'        :   'Python',
    'our_nosetests'     :   'Python',
    }

def build( thisdir, updir, script ):
    check_tested()

    main_script = script
    main_name   = main_script.replace( '.py', '' )

    bitness = platform.architecture()[0]
    if bitness == '64bit':
        bitness_name    = 'x64'
        # assert( updir.find( 'x64' ) >= 0 )
    elif bitness == '32bit':
        bitness_name    = 'x86'
        # assert( updir.find( 'x86' ) >= 0 )
    else:
        assert( False )

    print( 'Targeted architecture bitness seems to be consistent.' )

    ret = call( 'upx -V > nul', shell=True )
    if ret == 0:
        print( 'UPX is available.' )
        upx_is_avalable = True
    else:
        print( 'UPX is not available.' )
        upx_is_avalable = False

    built_dir = './dist/%s' % ( main_name )
    temp_dirs = [ 'build', 'dist', 'temp' ]
    for dir_ in temp_dirs:
        if os.path.exists( dir_ ):
            shutil.rmtree( dir_ )

    if main_name == 'executables':
        ret = call( 'pyinstaller %s --onedir' % ( main_script ), shell=True )
    else:
        class_name = class_name_dict[main_name]
        if main_name in [ 'autorg_kek', 'window_size_tester', 'python', 'pyinstaller-test' ]:
            windowed = ''
        else:
            windowed = '--windowed'
        command = 'pyinstaller %s --onedir %s --icon=../lib/%s/%s.ico' % ( main_script, windowed, class_name, main_name )
        print( command )
        ret = call( command, shell=True )

    assert( ret == 0 )
    assert( os.path.exists( built_dir ) )
    """
        mannually execute pyinstaller
        and see if "failed to "failed to create process" is shown.
    """

    python_install_dir = os.path.dirname( sys.executable )

    print( "Python is installed in '%s'" % ( python_install_dir ) )

    import numpy.core
    numpy_core_dir = os.path.dirname( numpy.core.__file__ )
    print( "numpy.core is in '%s'" % ( numpy_core_dir ) )

    if upx_is_avalable:
        working_dll_dir = 'temp'
        clear_dirs_with_retry( [ working_dll_dir ] )
    else:
        working_dll_dir = numpy_core_dir

    existing_dll_names = []
    dll_names = [ 'libiomp5md.dll', 'mkl_avx2.dll', 'mkl_core.dll', 'mkl_def.dll', 'mkl_intel_thread.dll' ]
    for name in dll_names:
        dll_path = os.path.join( numpy_core_dir, name )
        if os.path.exists( dll_path ):
            existing_dll_names.append( name )
            if upx_is_avalable:
                print( 'Compressing %s with UPX...' % ( name ) )
                working_dll_path = os.path.join( working_dll_dir, name )
                shutil.copy2( dll_path, working_dll_path )
                ret = call( 'upx %s' % ( working_dll_path ) )
                assert( ret == 0 )

    for name in existing_dll_names:
        shutil.copy2( os.path.join( working_dll_dir, name ), built_dir )

    if upx_is_avalable:
        vs140_dll_dir   = 'C:/Program Files (x86)/Microsoft Visual Studio 14.0/Common7/IDE/Remote Debugger/' + bitness_name
        pyrhonwin_dir   = python_install_dir + '/Lib/site-packages/pythonwin'
        badly_compressed_dlls = [
            os.path.join( numpy_core_dir,       'concrt140.dll'     ),
            os.path.join( numpy_core_dir,       'msvcp140.dll'      ),
            os.path.join( numpy_core_dir,       'vcamp140.dll'      ),
            os.path.join( numpy_core_dir,       'vccorlib140.dll'   ),
            os.path.join( numpy_core_dir,       'vcomp140.dll'      ),
            os.path.join( python_install_dir,   'vcruntime140.dll'  ),
            os.path.join( vs140_dll_dir,        'ucrtbase.dll'      ),
            os.path.join( pyrhonwin_dir,        'mfc140u.dll'       ),
            ]

        for path in badly_compressed_dlls:
            name = os.path.split( path )[-1]
            target_path = os.path.join( built_dir, name )
            if not os.path.exists( target_path ):
                continue
            upxname = name.replace( '.dll', '-upx.dll' )
            rename_path = os.path.join( built_dir, upxname )
            os.rename( target_path, rename_path )
            os.remove( rename_path  )
            shutil.copy2( path, target_path )
            print( 'copied from original to' + target_path )

    dist_dir = os.path.join( updir, main_name )
    if os.path.exists( dist_dir ):
        shutil.rmtree( dist_dir )

    rename_with_retry( built_dir, dist_dir )

    for dir_ in temp_dirs:
        if os.path.exists( dir_ ):
            shutil.rmtree( dir_ )

    os.remove( main_script.replace( '.py', '.spec' ) )

    if upx_is_avalable:
        with_phrase = ' with UPX compress'
    else:
        with_phrase = ''
    print( 'The distribution directory has been successfully built%s.' % ( with_phrase ) )

arc_dict = { '32bit':'x86', '64bit':'x64' }

def split_application( app=None, out_folder=None ):
    if app is not None and app != 'autorg_kek':
        return

    import AutorgKek
    from AppVersion import get_com_version_string
    vstr = get_com_version_string()
    def sub_func(m):
        global arch
        arch = arc_dict[m.group(2)]
        return m.group(1) + '-' + arch

    ver = re.sub( r'(autorg_kek\s+\d+\.\d+\.\d+).+\s+(\d+bit).+', sub_func, vstr ).replace('.', '_').replace(' ', '-')
    print( __file__ )
    assert( __file__.find( arch ) > 0 )
    print( ver, arch )
    path = __file__.replace( '\\', '/' )
    org_folder  = '/'.join( path.split( '/' )[0:-3] )
    up_folder   = '/'.join( path.split( '/' )[0:-4] )

    if out_folder is None:
        out_folder = up_folder + '/' + ver

    print( 'org_folder=', org_folder )
    print( 'out_folder=', out_folder )

    clear_dirs_with_retry( [ out_folder ] )

    for file in [ 'autorg_kek.bat', 'autorg_kek_client.bat' ]:
        print( 'Copying ' + file )
        src_path = org_folder + '/' + file
        tgt_path = out_folder + '/' + file
        shutil.copy2( src_path, tgt_path )

    # for subfolder in [ 'build', 'glue_devel', 'lib/KekLib', 'lib/AutorgKek', 'test_autorg_kek', 'executables' ]:
    for subfolder in [ 'build', 'glue_devel', 'lib/KekLib', 'lib/AutorgKek', 'test_autorg_kek' ]:
        print( 'Copying ' + subfolder )
        src_path = org_folder + '/' + subfolder
        tgt_path = out_folder + '/' + subfolder
        shutil.copytree( src_path, tgt_path )

    # remove/build/*
    for node in [   'build/__pycache__',
                    'build/split.py',
                    'build/serial_analyzer.py',
                    'build/executables.py',
                    'build/synthesizer.py',
                    'executables/synthesizer*',
                    'executables/serial_analyzer*',
                ]:
        tgt_path = out_folder + '/' + node

        if tgt_path.find('*') > 0:
            # print( 'glob ' + tgt_path )
            files = glob.glob(tgt_path)
            for f in files:
                print( 'Removing ' + f )
                os.remove( f )
            continue

        if not os.path.exists( tgt_path ):
            continue

        print( 'Removing ' + tgt_path )
        if os.path.isdir(tgt_path):
            shutil.rmtree( tgt_path )
        else:
            os.remove( tgt_path )

    # 
    tgt_dir = out_folder + '/lib/CommandMain'
    mkdirs_with_retry( tgt_dir )
    ini_path = tgt_dir + '/__init__.py'
    fh = open( ini_path, 'wb' )
    code = '\n'.join( [
        '"""',
        '    trick to make it backward compatible',
        '"""',
        'import KekLib',
        'import AutorgKek',
        'from AutorgKek.CommandMain  import CommandMain, autorg_func',
        '',
        ] )
    fh.write( code.encode() )
    fh.close()



# coding: utf-8
"""
    EasyScraper.py

    Copyright (c) 2021, Masatsuyo Takahashi, KEK-PF
"""
import requests
from bs4 import BeautifulSoup

class EasyScraper:
    def __init__(self, url):
        self.parser = self.get_page_parser(url)

    def get_page_parser(self, url):
        response = requests.get(url)
        print('response=', response)
        page_html = response.text
        soup = BeautifulSoup(page_html, "html.parser")
        return soup

    def get_element(self, tag, key_text):
        ret_text = None
        for k, element in enumerate(self.parser.findAll(tag)):
            if element.text.find(key_text) >= 0:
                ret_text = element.text
                break
        return ret_text



"""
    CompareModules.py

    Copyright (c) 2023, Masatsuyo Takahashi, KEK-PF
"""
import os
import sys
import re
import subprocess
from packaging import version
import ctypes

# https://stackoverflow.com/questions/1026431/cross-platform-way-to-check-admin-rights-in-a-python-script-under-windows
try:
    is_admin = os.getuid() == 0
except AttributeError:
    is_admin = ctypes.windll.shell32.IsUserAnAdmin() != 0

assert is_admin

this_python_exe = sys.executable

this_dir = os.path.dirname(os.path.abspath( __file__ ))
home_dir = os.path.abspath(os.path.join(this_dir, "..\\..\\"))
executables_dir = os.path.join(home_dir, "embeddables")
embed_pathon_exe = os.path.join(executables_dir, "python.exe")

assert os.path.exists(embed_pathon_exe)

requirements_txt = os.path.join(home_dir, "build\\build_embeddables\\requirements.txt")

truncate_re = re.compile(r"\s+#?.*")
version_re = re.compile(r"\bVersion:\s(\S+)\s")

with open(requirements_txt) as fh:
    for line in fh:
        if line[0] == "#":
            continue

        name = re.sub(truncate_re, '', line)
        print("Comparing", name)

        versions = []
        for python_exe in this_python_exe, embed_pathon_exe:
            result = subprocess.run([python_exe, "-m", "pip", "show", name], capture_output=True)
            stdout = result.stdout.decode()
            m = version_re.search(stdout)
            if m:
                versions.append(m.group(1))

        if len(versions) == 2 and versions[0] == versions[1]:
            print("ok", versions[0])
        else:
            print("different", versions)
            # synchronize to the later version
            if version.parse(versions[0]) < version.parse(versions[1]):
                ver = versions[1]
                python_exe = this_python_exe
            else:
                ver = versions[0]
                python_exe = embed_pathon_exe
            versioned_name = "%s==%s" % (name, ver)
            print('"%s" -m pip install %s' % (python_exe, versioned_name))
            result = subprocess.run([python_exe, "-m", "pip", "install", versioned_name], capture_output=True)
            if result.returncode == 0:
                print("ok")
            else:
                stdout = result.stdout.decode()
                print(stdout)
                stderr = result.stderr.decode()
                print(stderr)



"""
Global variable to hold command line arguments.
"""
from argparse import ArgumentParser

sys_args = None

def parse_sys_args():
    """
    Parse command line arguments and set global sys_args variable.
    """
    global sys_args

    ap = ArgumentParser()
    ap.add_argument('--devel', action='store_true', help='Enable development mode')
    ap.add_argument('--debug', action='store_true', help='Enable debug mode')
    sys_args = ap.parse_args()


"""
    UnifiedDecompResultTest.py

    Copyright (c) 2020,2025, SAXS Team, KEK-PF
"""
import logging
from matplotlib.patches import Rectangle

def plot_decomp_results(results, editor_ranges, use_debug_plot=False):
    """
    Plot the decomposition results using matplotlib.
    If use_debug_plot is True, use molass_legacy.KekLib.DebugPlot for plotting.
    """
    print("plot_decomp_results: editor_ranges=", editor_ranges)
    if use_debug_plot:
        import molass_legacy.KekLib.DebugPlot as plt
    else:
        import matplotlib.pyplot as plt

    def plot_decomp_results_impl():
        num_results = len(results)
        fig, axes = plt.subplots(ncols=num_results, figsize=(6 * num_results, 5))
        if num_results == 1:
            axes = [axes]
        for ax, result in zip(axes, results):
            x = result.x
            y = result.y
            ax.set_title("Decomposition Results")
            ax.plot(x, y)
            ymin, ymax = ax.get_ylim()
            ax.set_ylim(ymin, ymax)

            for list_ in editor_ranges:
                for f, t in list_:
                    p = Rectangle(
                            (f, ymin),  # (x,y)
                            t - f,   # width
                            ymax - ymin,    # height
                            facecolor   = 'cyan',
                            alpha       = 0.2,
                        )
                    ax.add_patch(p)

            for rec in result.opt_recs:
                func = rec[1]
                cy = func(x)
                ax.plot(x, cy, ":")

            axt = ax.twinx()
            axt.grid(False)
            for rec in result.opt_recs_uv:
                func = rec[1]
                cy = func(x)
                axt.plot(x, cy, "o", alpha=0.5)

        fig.tight_layout()
        plt.show()

    if use_debug_plot:
        with plt.Dp():
            plot_decomp_results_impl()
    else:
        plot_decomp_results_impl()

def unit_test(caller):
    from importlib import reload
    import UnifiedDecompResult
    reload(UnifiedDecompResult)
    from UnifiedDecompResult import UnifiedDecompResult

    editor = caller.dialog.get_current_frame()  # get editor because caller.editor is not updated by "Change Model"

    old_result = editor.decomp_result
    print("unit_test")
    # result = copy.deepcopy(old_result)
    result = UnifiedDecompResult(
                xray_to_uv=old_result.xray_to_uv,
                x_curve=old_result.x_curve,
                x=old_result.x,
                y=old_result.y,
                opt_recs=old_result.opt_recs,
                max_y_xray=old_result.max_y_xray,
                model_name=old_result.model_name,
                decomposer=old_result.decomposer,
                uv_y=old_result.uv_y,
                opt_recs_uv=old_result.opt_recs_uv,
                max_y_uv=old_result.max_y_uv,
                nresid_uv=old_result.nresid_uv,
                global_flag=old_result.global_flag,
                )

    logger = logging.getLogger(__name__)

    print("model_name=", result.model_name)
    result.remove_unwanted_elements()

    control_info = result.get_range_edit_info(logger=logger, debug=False)
    editor_ranges = control_info.editor_ranges
    print("editor_ranges=", editor_ranges)
    print("select_matrix=", control_info.select_matrix)
    print("top_x_list=", control_info.top_x_list)

    flags = result.identify_ignorable_elements()
    print("flags=", flags)

    plot_decomp_results([old_result, result], editor_ranges, use_debug_plot=True)



# coding: utf-8
"""

    ファイル名：   RequiredPackages.py

    処理内容：

        必須パッケージのインストール

    Copyright (c) 2016-2023, Masatsuyo Takahashi, KEK-PF
"""

from __future__ import division, print_function, unicode_literals
import sys
import os
import platform
import shutil
import re
import packaging    # to check, for it is required to build distributions
from subprocess import call
from BasicUtils import clear_dirs_with_retry, rename_with_retry, home_dir

required_packages = [
    'numpy',
    'matplotlib',
    'scipy',
    'packaging',
    'nose',
    'coverage',
    'statsmodels',
    # 'lmfit',
    'sklearn',
    'pymc',
    'seaborn',
    'screeninfo',
    'pyperclip',
    'openpyxl',
    'win32com',
    'pythoncom',
    'packaging',
    ]

site_url = 'http://www.lfd.uci.edu/~gohlke/pythonlibs/'

python_version_short = '%d%d' % sys.version_info[0:2]

bitness_str_map = {
    '32bit' : 'win32',
    '64bit' : 'win_amd64',
    }

bitness_str = bitness_str_map.get( platform.architecture()[0] )
if not bitness_str: assert( False )

postfix = '-cp%s-cp%sm-%s.whl' % ( python_version_short, python_version_short, bitness_str )

print( 'postfix=', postfix )

wheel_file_map = {
    # Caveate! : web page may include non-ascii characters.
    # e.g. '‑'(U+2011)s in 'numpy‑1.9.3+mkl‑cp35‑none‑win_amd64.whl'
    # Do not simply copy and paste!
    'numpy'         : 'numpy-1.11.2+mkl{0}'  .format( postfix ),
    'matplotlib'    : 'matplotlib-1.5.3{0}' .format( postfix ),
    'pandas'        : 'pandas-0.17.1{0}'    .format( postfix ),
    'scipy'         : 'scipy-0.16.1{0}'     .format( postfix ),
    }

to_be_installed_packages = []

def current_platform():
    return ( platform.python_version(), platform.system(), platform.architecture()[0] )

def check_tested():
    current_platform_ = current_platform()

    print( 'Current platform is python %s on %s %s.' % ( current_platform_ ) )

    tested_platforms = [
        ( '3.7.4rc1', 'Windows', '64bit' ),
        ( '3.7.3', 'Windows', '64bit' ),
        ( '3.7.2', 'Windows', '64bit' ),
        ( '3.7.1', 'Windows', '64bit' ),
        ( '3.7.0', 'Windows', '64bit' ),
        ( '3.6.8', 'Windows', '64bit' ),
        ( '3.6.7', 'Windows', '64bit' ),
        ( '3.6.6', 'Windows', '64bit' ),
        ( '3.6.5', 'Windows', '64bit' ),
        ( '3.6.3', 'Windows', '64bit' ),
        ( '3.6.2', 'Windows', '64bit' ),
        ( '3.6.1', 'Windows', '64bit' ),
        ( '3.5.2', 'Windows', '64bit' ),
        ( '3.5.1', 'Windows', '64bit' ),
        ( '3.5.0', 'Windows', '64bit' ),
        ( '3.5.2', 'Windows', '32bit' ),
        ( '3.5.0', 'Windows', '32bit' ),
        ]

    assert( current_platform_ in tested_platforms )

def requirements_ok():

    for package in required_packages:
        print( 'importing', package )
        try:
            exec( 'import ' + package )
        except:
            print( '%s is not yet installed.' % ( package ) )
            # print( sys.exc_info[0] )
            to_be_installed_packages.append( package )

    if len( to_be_installed_packages ) == 0:
        print( 'required modules seem to be installed.' )
        return True

    # install_packages()

    return False

def modification_ok():
    python_install_dir = os.path.dirname( sys.executable )
    # print( python_install_dir )
    pyinstaller_scrips = [
        'pyi-archive_viewer-script.py',
        'pyi-bindepend-script.py',
        'pyi-grab_version-script.py',
        'pyi-makespec-script.py',
        'pyi-set_version-script.py',
        'pyinstaller-script.py'
        ]
    first_line_modified_re = re.compile( '^".+"$' )
    for script in pyinstaller_scrips:
        path = os.path.join( python_install_dir, 'Scripts', script )
        # print( path )
        f = open( path )
        first_line = f.readline()
        f.close()
        # print( first_line )
        if first_line_modified_re.match( first_line ):
            print( '"%s" is not modified.' % path )
            return False

    print( 'pyinstaller scripts seem to be correctly modified.' )

    path = os.path.join( python_install_dir, 'Lib', 'site-packages', 'statsmodels', '__init__.py' )
    import_line_modified_re = re.compile( '^from\s+(\w*)\.tools\.sm_exceptions' )

    # statsmodels modification is still required as of pyinstaller 3.3.
    statsmodels_is_modified = False
    f = open( path )
    for i, line in enumerate( f.readlines() ):
        if i > 12:
            break
        # print( line )
        m = import_line_modified_re.match( line )
        if m:
            name = m.group(1)
            if name == 'statsmodels':
                statsmodels_is_modified = True
            else:
                print( 'WARNING: statsmodels is not modified.' )
    f.close()
    if statsmodels_is_modified:
        print( 'statsmodels seems to be correctly modified.' )
    # return statsmodels_is_modified
    return True

def install():
    check_tested()

    assert( platform.system() == 'Windows' )
    try:
        import win32api
    except:
        print( "Can't check admin privileges." )
        print( "If you are sure, comment out the following 'exit()' and retry." )
        # exit()

    ret = call( "pip -V", shell=True )
    if ret == 0:
        print( 'pip is available.' )
    else:
        print( 'Install pip manually.' )
        exit()

    """
    # TODO: download automation
    'wget' is not enough because we can't get target file url.

    print( 'Installing wget' )
    ret = call( "pip install wget" )
    if ret == 0:
        pass
    else:
        print( "Can't install wget. Please make sure to get admin privileges." )
        exit()
    """

    download_dir = os.path.join( home_dir(), 'Downloads' )
    # print( 'download_dir=', download_dir )

    num_skipped_packages = 0
    num_installed_packages = 0
    for package in required_packages:
        print( 'Installing %s' % ( package ) )
        wheel_file = wheel_file_map.get( package )
        # print( 'wheel_file=%s' % ( wheel_file) )

        if wheel_file == None:
            print( 'We will try to simply pip install {0}'.format( package ))
            ret = call( 'pip install {0}'.format( package ), shell=True )
            assert( ret == 0 )
        else:
            wheel_path = os.path.join( download_dir, wheel_file )
            # print( 'wheel_path=', wheel_path )

            if os.path.exists( wheel_path ):
                print( '%s has been downloaded, ok.' % ( wheel_file ) )
            else:
                print( '%s has not yet been downloaded.' % ( wheel_file ) )
                print( 'Skip installing {0}.'.format( package ) )
                num_skipped_packages += 1
                continue

            ret = call( 'pip install {0}'.format( wheel_path ), shell=True )
            assert( ret == 0 )
        num_installed_packages += 1

    assert( num_skipped_packages == 0 )
    assert( num_installed_packages == len( required_packages ) )

    print( 'All required packages have been successfully installed.' )



# coding: utf-8
"""
    SakuraUtils.py

    Copyright (c) 2021, Masatsuyo Takahashi, KEK-PF
"""
import os
import shutil

def clean_all_skrolds():
    this_dir, _ = os.path.split(os.path.abspath(__file__))
    root_dir = os.path.abspath(this_dir + '/../..')
    print(root_dir)

    for root, dirs, files in os.walk(root_dir):
        # print("root=", root)
        for file in files:
            if file.find('.skrold') > 0:
                path = os.path.join(root, file)
                print("removing", path)
                os.remove(path)



# coding: utf-8
"""
    Chrome.py

    Copyright (c) 2021-2022, Masatsuyo Takahashi, KEK-PF
"""
import os
import re
from .EasyScraper import EasyScraper

def get_installed_chrome_version():
    """
    cf. https://stackoverflow.com/questions/50880917/how-to-get-chrome-version-using-command-prompt-in-windows
    """
    dir_name_re = re.compile(r"^([\.\d]+)$")
    version = None
    for path in [   r"C:\Program Files\Google\Chrome\Application",
                    r"C:\Program Files (x86)\Google\Chrome\Application"]:
        if not os.path.exists(path):
            continue
        for line in os.listdir(path):
            m = dir_name_re.match(line)
            if m:
                version = m.group(1)
                break
    return version

def get_available_driver_version(chorme_version=None):
    if chorme_version is None:
        chorme_version = get_installed_chrome_version()

    url = "https://pypi.org/project/chromedriver-binary/#history"
    key_text = '.'.join(chorme_version.split('.')[0:3])
    es = EasyScraper(url)
    text = es.get_element('a', key_text)
    version_re = re.compile("(%s)" % (key_text.replace('.', '\\.') + '\\S+'))
    m = version_re.search(text)
    return m.group(1)



# coding: utf-8
"""
    PyCashe.py

    Copyright (c) 2020, Masatsuyo Takahashi, KEK-PF
"""
import os
import shutil

def clean_all_pycashes():
    this_dir, _ = os.path.split(os.path.abspath(__file__))
    root_dir = os.path.abspath(this_dir + '/..')
    print(root_dir)

    for root, dirs, files in os.walk(root_dir):
        # print("root=", root)
        for dir_ in dirs:
            # print("\tdir_=", dir_)
            if dir_ == "__pycache__":
                path = os.path.join(root, dir_)
                print("removing", path)
                shutil.rmtree(path)



"""
    UnofficialSite.py

    Copyright (c) 2020-2023, Masatsuyo Takahashi, KEK-PF
"""
import sys
import os
import platform
import re
import requests
import getpass
from time import sleep
from bs4 import BeautifulSoup
import glob
from shutil import copy
"""
learned at
1) https://qiita.com/Azunyan1111/items/b161b998790b1db2ff7a
2) https://qiita.com/memakura/items/20a02161fa7e18d8a693

    ChromeDriverをpipでバージョン指定してインストールする手順
        https://qiita.com/ten_to_ten/items/d0b39f7d1da1d2627d1b

    pip install chromedriver-binary==91.0.4472.101
"""
from selenium import webdriver
from selenium.webdriver.chrome.options import Options
import chromedriver_binary
from KnownPaths import get_downloads_folder

PAGE_URL = 'https://www.lfd.uci.edu/~gohlke/pythonlibs/'
"""
# MODULES_FROM_UO_SITE_WITH_LIST = ['pywin32', 'numpy', 'scipy', 'matplotlib', 'cupy',
MODULES_FROM_UO_SITE_WITH_LIST = ['pywin32', 'scipy', 'matplotlib',
                'pillow', 'kiwisolver', 'pandas', 'statsmodels', 'scikit-learn', 'fabio', 'lxml',
                'opencv']
# temporarily remove 'numba'
# temporarily remove 'llvmlite' to get llvmlite 0.34 for numba-0.51.2-cp39-cp39-win_amd64.whl

MODULES_FROM_UO_SITE_WITHOUT_LIST = ['future', 'asteval', 'et_xmlfile']
MODULES_FROM_UO_SITE = MODULES_FROM_UO_SITE_WITH_LIST + MODULES_FROM_UO_SITE_WITHOUT_LIST
"""
MODULES_FROM_UO_SITE = []

"""
crape data with beautifulsoup results in 404
https://stackoverflow.com/questions/41909065/scrape-data-with-beautifulsoup-results-in-404?rq=1
"""
HEADERS = {"User-Agent":"Mozilla/5.0"}

def get_downloadpage_parser(url):
    response = requests.get(url, headers=HEADERS)
    print('response=', response)
    page_html = response.text
    soup = BeautifulSoup(page_html, "html.parser")
    return soup

def get_downloadpage_driver(url):
    options = Options()
    # options.add_argument('--headless')        # --headless disturbes click download
    try:
        driver = webdriver.Chrome(options=options)
        driver.get(url)
    except:
        print(r"run as admin build\build_embeddables\tests\test_110_ChromeDriver.py")
        assert False
    # page_html = driver.page_source.encode('utf-8')
    return driver

class UnofficialSite:
    def __init__(self, NN):
        self.parser = get_downloadpage_parser(PAGE_URL)
        self.driver = get_downloadpage_driver(PAGE_URL)
        username = getpass.getuser()
        folder = get_downloads_folder().replace('Default', username)
        # note that the above replace does not take place unless is includes 'Default'

        print('folder=', folder)
        if not os.path.exists(folder):
            folder = folder.replace(username, 'Owner')
            print('folder=', folder)
            assert os.path.exists(folder)

        self.user_downloads_folder = folder

        self.rc_re = re.compile(r'rc\d*-')

        self.dl_a_element_dict = {}

        for name in MODULES_FROM_UO_SITE:
            if name == 'et_xmlfile':
                name_ = 'et-xmlfile'
            else:
                name_ = name
            a_element = self.parser.find(id=name_)
            print([name], a_element)
            self.dl_a_element_dict[name] = a_element

        if platform.architecture()[0].find('64') >= 0:
            self.uo_postfix = '‑cp{0}‑cp{0}‑win_amd64.whl'.format(NN)
        else:
            self.uo_postfix = '‑cp{0}‑cp{0}‑win32.whl'.format(NN)

    def is_ready_for(self, name):
        return self.dl_a_element_dict.get(name) is not None

    def get_a_element_text(self, name, spec=None):
        text = None
        start_a_element = self.dl_a_element_dict.get(name)

        if name in MODULES_FROM_UO_SITE_WITH_LIST:
            # find next ul element
            for j, ul_element in enumerate(start_a_element.find_all_next('ul')):
                # first-found ul_element
                break

            # find whl files
            for j, a_element in enumerate(ul_element.find_all_next('a')):
                text  = a_element.text
                print([j], text)
                if spec is not None:
                    if text.find(spec) < 0:
                        continue

                if text.find(self.uo_postfix) >= 0:
                    if self.rc_re.search(text.replace('‑','-')):
                        # skip release candidates.
                        # be careful not to be confused with hyphen-like chars.
                        print([j], 'skipping a release candidate.')
                        continue
                    else:
                        break
                if j > 20:
                    break
        else:
            for j, a_element in enumerate(start_a_element.find_all_next('a')):
                text  = a_element.text
                print([j], text)
                if text.find(name) >= 0 and text.find('py3'):
                    break
                if j > 20:
                    break

        print('get_a_element_text', text)
        return text

    def download(self, name, devel=False, spec=None):
        page_text = self.get_a_element_text(name, spec=spec)
        name_text = page_text.replace('‑', '-')

        file_path = os.path.join(self.user_downloads_folder, name_text)
        print('file_path=', file_path)
        if os.path.exists(file_path):
            print('there exists %s.' % file_path)
        else:
            # a_element = self.driver.find_element_by_link_text(page_text)
            # https://stackoverflow.com/questions/72754331/webdriver-object-has-no-attribute-find-element-by-link-text-selenium-scrip
            a_element = self.driver.find_element("link text", page_text)
            print(a_element)
            a_element.click()   # does not seem to work
            print("click done.")
            self.wait_for_download_complete(file_path)

        return file_path

    def wait_for_download_complete(self, file_path):
        dl_temp_file = file_path + '.crdownload'
        while (not os.path.exists(file_path) or os.path.exists(dl_temp_file)):
            sys.stdout.write('.')
            sys.stdout.flush()
            sleep(1)
        print('download complete in %s.' % file_path)

    def get_extra_step(self, name):
        if name == 'pywin32':
            extra_step = pywin32_extra_step
        else:
            extra_step = None

        return extra_step

def pywin32_extra_step(caller):
    target_folder = caller.emb_folder
    pywin32_dll_folder = os.path.join(caller.site_folder, 'pywin32_system32')
    for path in glob.glob(pywin32_dll_folder + r'\*.dll'):
        _, file = os.path.split(path)
        target_path = os.path.join(target_folder, file)
        if os.path.exists(target_path):
            print('%s exists, ok.' % file)
        else:
            print('copying %s.' % file)
            copy(path, target_folder)
        assert os.path.exists(target_path)



"""
    Embeddables.py

    Copyright (c) 2019-2024, Masatsuyo Takahashi, KEK-PF
"""
import sys
import os
import copy
import platform
import re
import zipfile
import subprocess
from shutil import copytree, copyfile
from time import sleep
import logging
from KekLib.StdoutLogger import StdoutLogger
from .Github import get_method
# from .SystemPath import add_to_path

REQUIRED_MINIMAL = ['requests', 'bs4', 'wget', 'selenium', 'chromedriver_binary']
EMBEDDABLES_FOLDER = 'embeddables'
TRY_ADDING_TO_PATH = False
SPECIFIED_VERSION = {
    "numpy": "1.26.4",
}

class DummyUnofficialSite:
    def __init__(self):
        pass

    def is_ready_for(self, name):
        return False

class Embeddables:
    def __init__(self, skip_install=False):
        self.logger = StdoutLogger("build.log")
        self.python00 = None
        self.python_version_re = re.compile(r'python-(\d+\.\d+\.\w+)')
        if not skip_install:
            self.install_required_minimum()
        self.emb_folder = EMBEDDABLES_FOLDER
        self.site_folder = os.path.join(self.emb_folder, r'Lib\site-packages')
        self.uo = DummyUnofficialSite()
        self.get_downloadpage_parser('https://www.python.org/downloads/windows/')

    def install_required_minimum(self):
        self.python_exe = sys.executable
        for name in REQUIRED_MINIMAL:
            print('trying to install %s.' % name)
            if self.already_installed(name):
                print('%s is already installed.' % name)
            else:
                if name == 'chromedriver_binary':
                    from .Chrome import get_available_driver_version
                    driver_version = get_available_driver_version()
                    name += '==%s' % driver_version
                args = [name]
                print('pip install', *args)
                ret = self.pip('install', *args)
                if ret.returncode != 0:
                    print(ret.stderr.decode())
                    assert False

    def get_this_ver(self):
        this_ver = 'Python %s' % platform.python_version()
        print('this_ver=', this_ver)
        return this_ver

    def clean(self):
        import shutil
        import glob
        for f in ['embeddables']:
            if os.path.exists(f):
                shutil.rmtree(f)
        for f in glob.glob('*.zip'):
            os.remove(f)

    def build(self, stable=True, requirements="requirements.txt"):
        """
        TODO: separate official site parsing
        """
        this_ver = self.get_this_ver()
        latest_ver = self.get_latest_version()
        if latest_ver == this_ver:
            want_ver = latest_ver
        else:
            print("Latest '%s' != this '%s'" % (latest_ver, this_ver))
            yn = input("Build from the latest version? ([y]/n) > ")
            if yn == "":
                yn = "y"
            print(yn)
            if yn[0] == "y":
                want_ver = latest_ver
            else:
                want_ver = this_ver

        self.build_ver = [int(v) for v in want_ver.split(' ')[-1].split('.')]
        print("build_ver=", self.build_ver)
        pair = self.get_latestrelease_top_pair()
        print('pair=', pair)

        url_pair = []
        for element in pair:
            url = self.get_embeddable_zipfile_url(element, want_ver)
            print('url=', url)
            url_pair.append(url)
            if url is not None:
                break

        # url = url_pair[0 if stable else 1]
        url = url_pair[0]
        file = url.split('/')[-1]

        if not os.path.exists('embeddables'):
            file = self.download(url, file)
            self.unzip(file)

        self.add_standard_dirs_to_path()
        self.prepare_python_exe()
        self.add_pip()
        if TRY_ADDING_TO_PATH:
            this_dir = os.path.dirname(os.path.abspath( __file__ ))
            script_path = os.path.join(os.path.join(this_dir, EMBEDDABLES_FOLDER), 'Scripts')
            print("script_path=", script_path)
            add_to_path(script_path)
        self.get_local_python_folder()
        self.copy_tkinter()
        self.copy_devel_folders()
        # self.prepare_unofficial_site()
        self.install_requirements(requirements)

    def prepare_python_exe(self):
        self.python_exe = os.path.join(self.emb_folder, 'python.exe')
        assert os.path.exists(self.python_exe)

    def get_downloadpage_parser(self, url):
        import requests
        from bs4 import BeautifulSoup
        response = requests.get(url)
        print('response=', response)
        page_html = response.text
        soup = BeautifulSoup(page_html, "html.parser")
        self.parser = soup
        return soup

    def get_latest_version(self):
        for k, h1 in enumerate(self.parser.findAll('h1')):
            # print([k], h1)
            if h1.text.find("Python Releases for Windows") >= 0:
                break

        for k, ul in enumerate(h1.find_all_next('ul')):
            # print([k], ul)
            break

        for k, li in enumerate(ul.find_all_next('li')):
            print([k], li)
            break

        python_version_re = re.compile(r"Latest Python 3 Release - (Python \S+)")
        m = python_version_re.search(li.text)
        if m:
            ret_ver = m.group(1)
        else:
            ret_ver = None
        return ret_ver

    def get_latestrelease_top_pair(self):
        pair = []
        for k, element in enumerate(self.parser.findAll('h2')):
            # print([k], element)
            pair.append(element)
        return pair

    def get_want_ver_title_element(self, top_element, want_ver):
        title_element = None
        for k, element in enumerate(top_element.find_all_next('a')):
            # print([k], element)
            if element.text.find(want_ver) >= 0 or k > 30:
                title_element = element
                break
        return title_element

    def get_embeddable_zipfile_url(self, top_element, want_ver):
        title_element = self.get_want_ver_title_element(top_element, want_ver)
        print('title_element=', title_element)

        version_num = re.sub(r"Python\s+(\S+)\s*", lambda m : m.group(1), want_ver)
        print("version_num='%s'" % version_num)

        if platform.architecture()[0].find('64') >= 0:
            key_name = 'embeddable package (64-bit)'
            old_key_name = 'x86-64 embeddable zip'
        else:
            key_name = 'embeddable package (32-bit)'
            old_key_name = 'x86 embeddable zip'

        url = None
        for name in [key_name, old_key_name]:
            for k, element in enumerate(title_element.find_all_next('a')):
                # print([k], element)
                if element.text.find(name) > 0:
                    url = element['href']
                    # print([k], element)
                    print([k], url)
                    if url.find(version_num) < 0:
                        continue

                    if self.python00 is None:
                        m = self.python_version_re.search(url)
                        if m:
                            v = m.group(1).split('.')
                            self.python00 = 'python' + v[0] + v[1]
                            print('python00=', self.python00)
                    break
            if url is None:
                continue
            else:
                break
        return url

    def download(self, url, out_file=None):
        import wget

        if out_file is None:
            filename = url.split('/')[-1]
            out_file = filename
        wget.download(url, out=out_file)
        return out_file

    def unzip(self, zipfile_):
        """
        see https://stackoverflow.com/questions/9813243/extract-files-from-zip-file-and-retain-mod-date-python-2-7-1-on-windows-7
        and improve this into a time-attributes-retaining one
        """
        with zipfile.ZipFile(zipfile_) as zh:
            zh.extractall(self.emb_folder)
        return self.emb_folder

    def add_standard_dirs_to_path(self):
        pth_file = self.emb_folder + '/' + self.python00 + '._pth'
        lines = []
        already_added = False
        with open(pth_file) as fh:
            for line in fh.readlines():
                lines.append(line)
                if line.find('import') == 0:
                    already_added = True
                    break
        if already_added:
            print('it seems that standard dirs have been already added.')
            return

        lines.insert(2, '.\\DLLs\n')
        lines.insert(2, '.\\Lib\n')
        lines.insert(2, '.\\Scripts\n')
        lines[-1] = lines[-1].replace('#', '')
        print(lines)
        new_pth_file = pth_file.replace('_pth', '_pth.tmp')
        with open(new_pth_file, 'w') as fh:
            fh.write(''.join(lines))

        bak_pth_file = pth_file.replace('_pth', '_pth.bak')
        os.rename(pth_file, bak_pth_file)
        os.rename(new_pth_file, pth_file)

    def python(self, *args):
        result = subprocess.run(
                [self.python_exe] + list(args),
                capture_output=True,
                )
        return result

    def pip(self, *args):
        result = self.python('-m', 'pip', *args)
        return result

    def add_pip(self):
        result = self.pip('show', 'pip')
        installed = result.returncode == 0
        if installed:
            print('it seems that pip has been installed.')
            return

        print('we must install pip.')
        file = self.download('https://bootstrap.pypa.io/get-pip.py')
        print("\nexecuting 'python %s'" % file)
        result = self.python(file)
        os.remove(file)

    def get_local_python_folder(self):
        self.local_python_folder, _ = os.path.split(sys.executable)
        print('local python is installed in "%s"' % self.local_python_folder)

    def copy_folders(self, folders):
        for folder in folders:
            src_folder = os.path.join(self.local_python_folder, folder)
            tgt_folder = os.path.join(self.emb_folder, folder)
            if os.path.exists(tgt_folder):
                print(tgt_folder, 'exists.')
            else:
                print('copying to', tgt_folder)
                copytree(src_folder, tgt_folder)

    def copy_tkinter(self):
        self.copy_folders(['tcl', 'Lib/tkinter', 'Lib/idlelib'])

        dll_files = ['_tkinter.pyd', 'tcl86t.dll', 'tk86t.dll']
        if self.build_ver >= [3,12,0]:
            dll_files += ['zlib1.dll']      # zlib1.dll is required from Python312
        src_dll_folder = os.path.join(self.local_python_folder, 'DLLs')
        tgt_dll_folder = os.path.join(self.emb_folder, 'DLLs')
        if not os.path.exists(tgt_dll_folder):
            os.makedirs(tgt_dll_folder)

        for file in dll_files:
            src_file = os.path.join(src_dll_folder, file)
            tgt_file = os.path.join(tgt_dll_folder, file)
            if os.path.exists(tgt_file):
                print(tgt_file, 'exists.')
            else:
                print('copying to', tgt_file)
                copyfile(src_file, tgt_file)

    def copy_devel_folders(self):
        self.copy_folders(['include', 'libs'])

    def already_installed(self, name):
        print('pip show', name)
        ret = self.pip('show', name)
        return ret.returncode == 0

    def install_requirements(self, requirements):
        """
        TODO:   check version compatibility other than patch numbers.
                e.g.,
                    3.1.2 ==> 3.1.3 ok
                    3.1.2 ==> 3.2.0 not ok, or be careful at least
        """
        truncate_re = re.compile(r'\s+#?.*')
        spec_re = re.compile(r'spec=(\S+)')
        with open(requirements) as fh:
            for line in fh.readlines():
                if line[0] == '#':
                    continue
                name = re.sub(truncate_re, '', line)
                print('trying to install %s.' % name)
                if self.already_installed(name):
                    print('%s is already installed.' % name)
                else:
                    method = get_method(name)
                    if method is None:
                        if self.uo.is_ready_for(name):
                            m = spec_re.search(line)
                            if m:
                                spec = m.group(1)
                            else:
                                spec = None
                            arg_name = self.uo.download(name, spec=spec)
                        else:
                            arg_name = name
                        
                        version = SPECIFIED_VERSION.get(arg_name)
                        if version is not None:
                            arg_name += "==%s" % version
                        print('pip install "%s"' % arg_name)
                        ret = self.pip('install', arg_name)
                        if ret.returncode != 0:
                            print(ret.stderr.decode())
                            assert False
                    else:
                        method()
                if self.uo.is_ready_for(name):
                    extra_step = self.uo.get_extra_step(name)
                    if extra_step is not None:
                        extra_step(self)

    def prepare_unofficial_site(self):
        from .UnofficialSite import UnofficialSite
        NN = self.python00.replace('python', '')
        self.uo = UnofficialSite(NN)



# coding: utf-8
"""
    Github.py

    Copyright (c) 2020-2021, Masatsuyo Takahashi, KEK-PF
"""
import os
import re
import subprocess

def where(name):
    result = subprocess.run(
            [r'C:\Windows\System32\where.exe', name],
            capture_output=True,
            )
    return result.stdout[:-2].decode()  # [:-2] is to remove \r\n

class Github:
    def __init__(self):
        self.git_exe = where('git')

    def git(self, *args):
        result = subprocess.run(
                [self.git_exe] + list(args),
                capture_output=True,
                )
        return result

    def clone(self, url_git):
        self.git('clone', url_git)

gh = None

def get_gh():
    global gh
    if gh is None:
        gh = Github()
    return gh

def download_pytweening(folder):
    if os.path.exists(folder):
        print('already exists')
    else:
        gh = get_gh()
        url = 'https://github.com/asweigart/pytweening.git'
        print('git clone ' + url)
        gh.clone(url)
        if os.path.exists(folder):
            print('ok')

def edit_pytweening():
    src = open('setup.py').read()
    new_src = re.sub(r"version=.+,", "version='1.0.3',", src)

    new_file = 'setup-modified.py'
    with open(new_file, 'w') as fh:
        fh.write(new_src)
    return new_file

def run_setup_install(python_exe, setup_py):
    print('%s setup.py install' % python_exe)
    result = subprocess.run(
            [python_exe, setup_py, 'install'],
            capture_output=True,
            )
    return result

EMBEDDABLES_PYTHON_EXE = r'..\embeddables\python.exe'

def install_pytweening():
    folder = 'pytweening'
    download_pytweening(folder)
    cwd = os.getcwd()
    os.chdir(folder)
    new_file = edit_pytweening()
    result = run_setup_install(EMBEDDABLES_PYTHON_EXE, new_file)
    # print(result)
    assert result.returncode == 0
    print('successfully installed pytweening')
    os.chdir(cwd)
    print(os.getcwd())

# METHODS_DICT = {'pytweening':install_pytweening}  # no longer required as of 2022-01-18
METHODS_DICT = {}

def get_method(k):
    return METHODS_DICT.get(k)
